                        // frac.cc
#include "frac.h"

#define ASSERT_YES
#undef EXCEPTION
#undef NOERROR_TRAP

// this algorithm doesn't alter _num and _denom signs
// precondition: _denom != 0
void fraction::reduce()
{
   long  N=_num, D=_denom, R;

   while ((R = N % D)) // != 0
   {
      N=D;
      D=R;
   }
   // MCD is D

   if (D!=1 && D!=-1)   // optional test
   {
      _num /= D;
      _denom /= D;
   }
}

ostream& operator<<(ostream& os, const fraction& f)
{
   return os << f._num << '/' << f._denom;   
}

istream& operator>>(istream& is, fraction& f)
{
   is >> f._num >> f._denom;
   f.ExceptionIfDenom0();
   f.reduce();
   return is;
}

fraction operator*(const fraction& f1, const fraction& f2)
{
   return fraction(f1._num*f2._num, f1._denom*f2._denom);
}

fraction operator/(const fraction& f1, const fraction& f2)
{
   return fraction(f1._num*f2._denom, f1._denom*f2._num);
}

bool operator<(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom < f1._denom*f2._num;
}

bool operator<=(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom <= f1._denom*f2._num;
}

bool operator>(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom > f1._denom*f2._num;
}

bool operator>=(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom >= f1._denom*f2._num;
}

bool operator==(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom == f2._num*f1._denom;
}

bool operator!=(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom != f2._num*f1._denom;
}

long mcm(long a, long b)
{
   long mcm;

   if (a<b)
   {
      long tmp=a;
      a=b;
      b=tmp;
   }

   mcm=a;
   while (mcm % b)
      mcm+=a;
   return mcm;
}

fraction operator+(const fraction& f1, const fraction& f2)
{
   long m=mcm(f1._denom, f2._denom);
   return fraction(m/f1._denom*f1._num + m/f2._denom*f2._num, m);
}

fraction operator-(const fraction& f1, const fraction& f2)
{
   long m=mcm(f1._denom, f2._denom);
   return fraction(m/f1._denom*f1._num - m/f2._denom*f2._num, m);
}
