                        // frac.cc

#include "frac.h"
#include <stdlib.h>
#include <ctype.h>
#include <iomanip.h>

#define ASSERT_YES
#undef EXCEPTION
#undef NOERROR_TRAP

template <class T>
static inline T max(T a, T b)
{
   return a>b? a : b;
}

static char numlen(long n)
{
   char len=0;
   do
      ++len;
   while (n/=10);
   return(len);
}

// this algorithm doesn't alter _num and _denom signs
// precondition: _denom != 0
void fraction::reduce()
{
   long  N=_num, D=_denom, R;

   while ((R = N % D)) // != 0
   {
      N=D;
      D=R;
   }
   // MCD is D

   if (D!=1 && D!=-1)   // optional test
   {
      _num /= D;
      _denom /= D;
   }
}

ostream& operator<<(ostream& os, fraction f)
{
   if (f._denom<0)
   {
      f._denom = 0-f._denom;
      f._num = 0-f._num;
   }
   os << f._num;
   if (f._denom != 1)
      os << '/' << f._denom;
   return os;
}

istream& operator>>(istream& is, fraction& f)
{
   int c;

   f._denom=1;

   if (!(is >> f._num))
      f._num=0;
   else
   {
      while (isspace(c=is.get()) && c!='\n')
         ;
      if (c!='/' && c!=':')
      {
         if (c!='\n') is.unget();
      }
      else if (is >> f._denom) // or (is >> f._denom).good()
      {
         f.ExceptionIfDenom0();
         f.reduce();
      }
   }
   return is;
}

fraction operator*(const fraction& f1, const fraction& f2)
{
   return fraction(f1._num*f2._num, f1._denom*f2._denom);
}

fraction operator/(const fraction& f1, const fraction& f2)
{
   return fraction(f1._num*f2._denom, f1._denom*f2._num);
}

bool operator<(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom < f1._denom*f2._num;
}

bool operator<=(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom <= f1._denom*f2._num;
}

bool operator>(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom > f1._denom*f2._num;
}

bool operator>=(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom >= f1._denom*f2._num;
}

bool operator==(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom == f2._num*f1._denom;
}

bool operator!=(const fraction& f1, const fraction& f2)
{
   return f1._num*f2._denom != f2._num*f1._denom;
}

long mcm(long a, long b)
{
   long mcm;

   if (a<b)
   {
      long tmp=a;
      a=b;
      b=tmp;
   }

   mcm=a;
   while (mcm % b)
      mcm+=a;
   return mcm;
}

fraction operator+(const fraction& f1, const fraction& f2)
{
   long m=mcm(f1._denom, f2._denom);
   return fraction(m/f1._denom*f1._num + m/f2._denom*f2._num, m);
}

fraction operator-(const fraction& f1, const fraction& f2)
{
   long m=mcm(f1._denom, f2._denom);
   return fraction(m/f1._denom*f1._num - m/f2._denom*f2._num, m);
}

void fraction::display(ostream& os=cout) const
{
   if (_denom==1)
   {
      os << _num << endl;
      return;
   }
   if (_denom==-1)
   {
      os << -_num << endl;
      return;
   }

   const char dash='-';

   char _numlen=numlen(_num),
        _denomlen=numlen(_denom),
        dashes=max(_numlen, _denomlen)+2,
        minus=_num*_denom > 0? 0 : 2;

   os << setw((dashes-_numlen+1)/2+_numlen+minus) << labs(_num) << endl;

   // print the possible minus sign
   if (minus > 0)
      os << "- ";

   // print dashes
   for (char i=0; i<dashes; i++)
      os.put(dash);

   os << endl;
   os << setw((dashes-_denomlen+1)/2+_denomlen+minus) << labs(_denom) << endl;
}

fraction fraction::operator^(long exp) const
{
   if (_denom==0 && exp<=0)
      ExceptionIfDenom0();

   long num, den;

   num=den=1;
   for (long i=0; i<labs(exp); i++)
   {
      num*=_num;
      den*=_denom;
   }

   if (exp<0)
      return fraction(den, num);
   return fraction(num, den);
}
