			// frac.h
#ifndef  _fraction_H
#define  _fraction_H

// choose from: ASSERT_YES (default), NOERROR_TRAP, EXCEPTION
#define ASSERT_YES

#include <iostream.h>   // for istream, ostream
#include <assert.h>     // for macro assert()

class fraction;
typedef fraction fract;
typedef fraction frac;

class fraction
{
   // binary operators //
   friend ostream& operator<<(ostream&, fraction);
   friend istream& operator>>(istream&, fraction&);

   friend fraction operator+(const fraction&, const fraction&);
   friend fraction operator+(const fraction& f, long n)
      { return fraction(f._num + n*f._denom, f._denom); }
   friend fraction operator+(long n, const fraction& f)
      { return f+n; }

   friend fraction operator-(const fraction&, const fraction&);
   friend fraction operator-(const fraction& f, long n)
      { return fraction(f._num - n*f._denom, f._denom); }
   friend fraction operator-(long n, const fraction& f)
      { return fraction(n*f._denom - f._num, f._denom); }

   friend fraction operator*(const fraction&, const fraction&);
   friend fraction operator*(const fraction& f, long n)
      { return fraction(f._num*n, f._denom); }
   friend fraction operator*(long n, const fraction& f)
      { return f*n; }

   friend fraction operator/(const fraction&, const fraction&);
   friend fraction operator/(const fraction& f, long n)
      { return fraction(f._num, f._denom*n); }
   friend fraction operator/(long n, const fraction& f)
      { return fraction(n*f._denom, f._num); }

   friend bool operator<(const fraction&, const fraction&);
   friend bool operator<(const fraction& f, long n)
      { return f._num < f._denom*n; }
   friend bool operator<(long n, const fraction& f)
      { return n*f._denom < f._num; }

   friend bool operator<=(const fraction&, const fraction&);
   friend bool operator<=(const fraction& f, long n)
      { return f._num <= f._denom*n; }
   friend bool operator<=(long n, const fraction& f)
      { return n*f._denom <= f._num; }

   friend bool operator>(const fraction&, const fraction&);
   friend bool operator>(const fraction& f, long n)
      { return f._num > f._denom*n; }
   friend bool operator>(long n, const fraction& f)
      { return n*f._denom > f._num; }

   friend bool operator>=(const fraction&, const fraction&);
   friend bool operator>=(const fraction& f, long n)
      { return f._num >= f._denom*n; }
   friend bool operator>=(long n, const fraction& f)
      { return n*f._denom > f._num; }

   friend bool operator==(const fraction&, const fraction&);
   friend bool operator==(const fraction& f, long n)
      { return f._num == n*f._denom; }
   friend bool operator==(long n, const fraction& f)
      { return f==n; }

   friend bool operator!=(const fraction&, const fraction&);
   friend bool operator!=(const fraction& f, long n)
      { return f._num != n*f._denom; }
   friend bool operator!=(long n, const fraction& f)
      { return f!=n; }

public:
   #ifdef EXCEPTION
      // exceptions //
      class DivisionBy0 {};
   #endif

   // constructors //
   fraction(): _num(0), _denom(1) {}
   fraction(long n, long d=1): _num(n), _denom(d)
      { ExceptionIfDenom0(); reduce(); }

   // accessors //
   long num() const { return _num; }
   long denom() const { return _denom; }
   void set_num(long n) { _num=n; reduce(); }
   void set_denom(long d)
      { _denom=d; ExceptionIfDenom0(); reduce(); }

   // others //
   void display(ostream& = cout) const; // nice way to display a fraction
   operator double( ) const { return (double)_num/_denom; } // type converter
   fraction abs( ) const { return fraction(labs(_num), labs(_denom)); }

   // unary operators //
   fraction operator+( ) const { return fraction(*this); }
   fraction operator-( ) const { return fraction(-_num, _denom); }

   // other binary operators //
   fraction operator^(long) const;

protected:
   long _num, _denom;

   void ExceptionIfDenom0() const
   {
      #ifdef EXCEPTION
         if (_denom==0) throw DivisionBy0( );
      #elif defined(NOERROR_TRAP)
         // don't trap this error!
      #else // ASSERT_YES
         assert(_denom != 0);
      #endif
   }

   void reduce();
};

#endif
