			// frac.h
#ifndef  _fraction_H
#define  _fraction_H

// choose from: ASSERT_YES (default), NOERROR_TRAP, EXCEPTION
#define EXCEPTION

#include <iostream.h>   // for istream, ostream
#include <assert.h>     // for macro assert()

class fraction;
typedef fraction fract;
typedef fraction frac;

typedef long fitype;
const char FITYPE_DIG = 9; // if long int's are 32 bit long

class fraction
{
   // binary operators //
   friend ostream& operator<<(ostream&, fraction);
   friend istream& operator>>(istream&, fraction&);

   friend const fraction operator+(const fraction&, const fraction&);
   friend const fraction operator+(const fraction& f, fitype n)
      { return fraction(f._num + n*f._denom, f._denom); }
   friend const fraction operator+(fitype n, const fraction& f)
      { return f+n; }

   friend const fraction operator-(const fraction&, const fraction&);
   friend const fraction operator-(const fraction& f, fitype n)
      { return fraction(f._num - n*f._denom, f._denom); }
   friend const fraction operator-(fitype n, const fraction& f)
      { return fraction(n*f._denom - f._num, f._denom); }

   friend const fraction operator*(const fraction&, const fraction&);
   friend const fraction operator*(const fraction& f, fitype n)
      { return fraction(f._num*n, f._denom); }
   friend const fraction operator*(fitype n, const fraction& f)
      { return f*n; }

   friend const fraction operator/(const fraction&, const fraction&);
   friend const fraction operator/(const fraction& f, fitype n)
      { return fraction(f._num, f._denom*n); }
   friend const fraction operator/(fitype n, const fraction& f)
      { return fraction(n*f._denom, f._num); }

   friend bool operator<(const fraction&, const fraction&);
   friend bool operator<(const fraction& f, fitype n)
      { return f._num < f._denom*n; }
   friend bool operator<(fitype n, const fraction& f)
      { return n*f._denom < f._num; }

   friend bool operator<=(const fraction&, const fraction&);
   friend bool operator<=(const fraction& f, fitype n)
      { return f._num <= f._denom*n; }
   friend bool operator<=(fitype n, const fraction& f)
      { return n*f._denom <= f._num; }

   friend bool operator>(const fraction&, const fraction&);
   friend bool operator>(const fraction& f, fitype n)
      { return f._num > f._denom*n; }
   friend bool operator>(fitype n, const fraction& f)
      { return n*f._denom > f._num; }

   friend bool operator>=(const fraction&, const fraction&);
   friend bool operator>=(const fraction& f, fitype n)
      { return f._num >= f._denom*n; }
   friend bool operator>=(fitype n, const fraction& f)
      { return n*f._denom > f._num; }

   friend bool operator==(const fraction&, const fraction&);
   friend bool operator==(const fraction& f, fitype n)
      { return f._num == n*f._denom; }
   friend bool operator==(fitype n, const fraction& f)
      { return f==n; }

   friend bool operator!=(const fraction&, const fraction&);
   friend bool operator!=(const fraction& f, fitype n)
      { return f._num != n*f._denom; }
   friend bool operator!=(fitype n, const fraction& f)
      { return f!=n; }
  
public:
   #ifdef EXCEPTION
      // exceptions //
      class DivisionBy0 {};
   #endif

   // constructors //
   fraction(): _num(0), _denom(1) {}
   fraction(fitype n, fitype d=1): _num(n), _denom(d) // type converter
      { ExceptionIfDenom0(); reduce(); }

   // accessors //
   const fitype& num() const { return _num; }
   const fitype& denom() const { return _denom; }
   void set_num(fitype n) { _num=n; reduce(); }
   void set_denom(fitype d)
      { _denom=d; ExceptionIfDenom0(); reduce(); }

   // others //
   void display(ostream& = cout) const; // nice way to display a fraction
   operator double( ) const { return (double)_num/_denom; } // type converter
   operator bool( ) const { return _num!=0; } // optional
   const fraction abs( ) const { return fraction(labs(_num), labs(_denom)); }
   void setfd(double d, char p=FITYPE_DIG); // SET fraction value From Double
   friend const fraction ffd(double d, char p=FITYPE_DIG); // type converter on demand

   // other binary operators //
   const fraction operator^(fitype) const;

   // unary operators //
   const fraction operator+( ) const { return fraction(*this); }
   const fraction operator-( ) const { return fraction(-_num, _denom); }

   const fraction& operator++( ) { _num+=_denom; return (*this); } // prefix
   const fraction operator++(int);                                 // postfix
   const fraction& operator--( ) { _num-=_denom; return (*this); } // prefix
   const fraction operator--(int);                                 // postfix

   const fraction& operator +=(const fraction&);
   const fraction& operator -=(const fraction&);
   const fraction& operator *=(const fraction&);
   const fraction& operator /=(const fraction&);

   const fraction& operator +=(fitype);
   const fraction& operator -=(fitype);
   const fraction& operator *=(fitype);
   const fraction& operator /=(fitype);

protected:
   fitype _num, _denom;

   void ExceptionIfDenom0() const
   {
      #ifdef EXCEPTION
         if (_denom==0) throw DivisionBy0( );
      #elif defined(NOERROR_TRAP)
         // don't trap this error!
      #else // ASSERT_YES
         assert(_denom != 0);
      #endif
   }

   void reduce();
};

#endif
