#include "frac.h"

int main()
{
   short n,    // system dimension
         step, // algorithm step counter
         k,    // row index
         i;    // column index

   cout << "* solve a squared non-singular linear system\n"
           "* with rational coefficients\n\n";
   cout << "system dimension=";
   cin >> n;

   // uses a GNU extension
   frac lsys[n][n+1],           // linear system
        *sol = new frac[n],     // solution vector
        l,                      // multiplier (lambda)
        temp;                   // a temporary fraction

   cout << "input system matrix by rows\n";
   for (k=0; k<n; ++k)
      for (i=0; i<=n; ++i)
         cin >> lsys[k][i];

   // 1. triangolarize
   for (step=0; step<n-1; ++step)
      for (k=step; k<n; ++k)       // k is row index
         // look for a non-zero element in column step, rows step..n-1
         if (lsys[k][step])
         {
            // found

            // swaps row step and k to place the pivot
            if (step != k) // if needed (optional)
               for (i=0; i<=n; i++)
               {
                  temp=lsys[step][i];
                  lsys[step][i]=lsys[k][i];
                  lsys[k][i]=temp;
               }

            // "triangolarize" this column
            for (k=step+1; k<n; ++k)       // we are reusing k
            {
               l=lsys[k][step]/lsys[step][step];

               for (i=0; i<=n; ++i)        // we are reusing i
                  lsys[k][i]-=l*lsys[step][i];
            }

            break;
         }

   // 2. back substitution
   try
   {
      for (k=n-1; k>=0; --k)
      {
         // solve the equation no k
         sol[k]=lsys[k][n];
         for (i=n-1; i>k; --i)
            sol[k]-=lsys[k][i]*sol[i];
         sol[k]/=lsys[k][k];    // or lsys[k][i]
      }
   }
   catch (frac::DivisionBy0)
   {
      cout << "sorry, can't handle singular systems\n";
      exit(0);
   };

   // 3. print solutions
   for (k=0; k<n; ++k)
   {
      cout << "\n#" << k+1 << ":\n";
      sol[k].display();
   }

   cout << "\nbye :)\n";

   exit(0);
}

